///////////////////////////////////////////////////////////////////
// Functy
// 3D graph drawing utility
//
// David Llewellyn-Jones
// http://www.flypig.co.uk
//
// Spring 2009
//
// GTK code inspired by GtkGLExt logo demo
// by Naofumi Yasufuku <naofumi@users.sourceforge.net>
///////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////
// Includes

#include "utils.h"

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkgl.h>
#include <gio/gio.h>

#include "functy.h"
#include "filesave.h"
#include "spherical.h"
#include "curve.h"

///////////////////////////////////////////////////////////////////
// Defines

#define UPDATE_TIMEOUT (10)
#define FULLSCREEN_BUTTONBARSHOW (8)
#define RADIUS_SCROLL (1)
#define FUNCTION_NEW_CARTESIAN "0"
#define FUNCTION_NEW_SPHERICAL "1"
#define FUNCTION_NEW_CURVE_RADIUS "1"
#define FUNCTION_NEW_CURVE_X "a"
#define FUNCTION_NEW_CURVE_Y "0"
#define FUNCTION_NEW_CURVE_Z "0"
#define ALPHA_NEW (0.9)
#define COLOUR_NEW_SHIFT (2.3)
#define RANGE_FORMAT "%g"
#define DEFAULT_FILENAME "Functions.xml"
#define DEFAULT_EXPORTMODELNAME "Model.ply"
#define RANGE_SCALEUP (1.08)
#define RANGE_SCALEDOWN (1.0/RANGE_SCALEUP)
#define SPHERICAL_XCENTRE (0.0)
#define SPHERICAL_YCENTRE (0.0)
#define SPHERICAL_ZCENTRE (0.0)
#define CURVE_XCENTRE (0.0)
#define CURVE_YCENTRE (0.0)
#define CURVE_ZCENTRE (0.0)

///////////////////////////////////////////////////////////////////
// Structures and enumerations

typedef enum _FUNCSCOL {
  FUNCSCOL_INVALID = -1,

  FUNCSCOL_FUNCDATA,
  FUNCSCOL_FUNCTION,

  FUNCSCOL_NUM
} _FUNCSCOL;

///////////////////////////////////////////////////////////////////
// Global variables

///////////////////////////////////////////////////////////////////
// Function prototypes

static void Realize (GtkWidget * psWidget, gpointer psData);
static gboolean ConfigureEvent (GtkWidget * psWidget, GdkEventConfigure * psEvent, gpointer psData);
static gboolean ExposeEvent (GtkWidget * psWidget, GdkEventExpose * psEvent, gpointer psData);
static gboolean ButtonPressEvent (GtkWidget * psWidget, GdkEventButton * psEvent, gpointer * psData);
static gboolean ButtonReleaseEvent (GtkWidget * psWidget, GdkEventButton * psEvent, gpointer * psData);
//static gboolean MotionNotifyEvent (GtkWidget * psWidget, GdkEventMotion * psEvent, gpointer psData);
static gboolean ScrollEvent (GtkWidget * psWidget, GdkEventScroll * psEvent, gpointer psData);
static gboolean KeyPressEvent (GtkWidget * psWidget, GdkEventKey * psEvent, gpointer psData);
static gboolean KeyReleaseEvent (GtkWidget * psWidget, GdkEventKey * psEvent, gpointer psData);
static gboolean Timeout (gpointer psData);
static void TimeoutAdd (FunctyPersist * psFunctyData);
static void TimeoutRemove (FunctyPersist * psFunctyData);
static gboolean MapEvent (GtkWidget * psWidget, GdkEventAny * psEvent, gpointer psData);
static gboolean UnmapEvent (GtkWidget * psWidget, GdkEventAny * psEvent, gpointer psData);
static gboolean VisibilityNotifyEvent (GtkWidget * psWidget, GdkEventVisibility * psEvent, gpointer psData);
void ToggleFullScreenWindow (FunctyPersist * psFunctyData);
void ToggleButtonBar (FunctyPersist * psFunctyData);
static gboolean PlotFunctionPress (GtkWidget * psWidget, gpointer psData);
static gboolean PlotFunctionApply (GtkWidget * psWidget, gpointer psData);
static gboolean ToggleSpinPress (GtkToggleButton * psWidget, gpointer psData);
static gboolean ToggleInvertPress (GtkToggleButton * psWidget, gpointer psData);
static gboolean ToggleFullscreenPress (GtkToggleButton * psWidget, gpointer psData);
static gboolean ToggleDrawAxesPress (GtkToggleButton * psWidget, gpointer psData);
static gboolean ToggleWireframePress (GtkToggleButton * psWidget, gpointer psData);
static gboolean ViewPress (GtkWidget * psWidget, gpointer psData);
void EditFunctionRowDoubleClick (GtkTreeView * psTreeView, GtkTreePath * psPath, GtkTreeViewColumn * psColumn, gpointer psData);
static gboolean EditFunctionPress (GtkWidget * psWidget, gpointer psData);
void EditFunction (GtkTreeModel * psTreeModel, GtkTreeIter * psTreeIter, FunctyPersist * psFunctyData);
void PopulateFunctionList (GtkListStore * psFunctionListModel, FunctyPersist * psFunctyData);
void AddFunction (FuncPersist * psFuncData, GtkListStore * psFunctionListModel);
void FunctionCellDataFunc (GtkTreeViewColumn * psTreeColumn, GtkCellRenderer * psCell, GtkTreeModel * psTreeModel, GtkTreeIter * psIter, gpointer psData);
void SetColourButton (FunctyPersist * psFunctyData);
void SetColourButtonCallback (GtkColorButton * psWidget, gpointer psData);
void SetViewWindowState (FunctyPersist * psFunctyData);
void SetRangeWindowState (FunctyPersist * psFunctyData);
static gboolean RangeSetPress (GtkWidget * psWidget, gpointer psData);
static gboolean RangeSetApply (GtkWidget * psWidget, gpointer psData);
gboolean RangeMinChange (GtkWidget *psWidget, GdkEventKey * psEvent, gpointer psData);
gboolean RangeMaxChange (GtkWidget *psWidget, GdkEventKey * psEvent, gpointer psData);
void GetNextColour (double * pfRed, double * pfGreen, double * pfBlue, double * pfAlpha, FunctyPersist * psFunctyData);
static void ClearAllFunctions (FunctyPersist * psFunctyData);
static gboolean LoadFilePress (GtkWidget * psWidget, gpointer psData);
static gboolean SaveFilePress (GtkWidget * psWidget, gpointer psData);
static gboolean ExportFilePress (GtkWidget * psWidget, gpointer psData);
void PauseAnimationModal (bool boPauseModal, FunctyPersist * psFunctyData);
void NewCartesianFunction (FunctyPersist * psFunctyData);
void NewSphericalFunction (FunctyPersist * psFunctyData);
void NewCurveFunction (FunctyPersist * psFunctyData);
static gboolean AddCartesianPress (GtkColorButton * psWidget, gpointer psData);
static gboolean AddSphericalPress (GtkColorButton * psWidget, gpointer psData);
static gboolean AddCurvePress (GtkColorButton * psWidget, gpointer psData);
gchar * GetTitleFilename (char const * szFilename);
void SetMainWindowTitle (char const * szFilename, FunctyPersist * psFunctyData);

///////////////////////////////////////////////////////////////////
// Function definitions

FunctyPersist * NewFunctyPersist (void) {
  FunctyPersist * psFunctyData;

  psFunctyData = g_new0 (FunctyPersist, 1);

  psFunctyData->psVisData = NULL;
  psFunctyData->psDrawingArea = NULL;
  psFunctyData->boTimeoutContinue = FALSE;
  psFunctyData->TimeoutID = 0;
  psFunctyData->psFuncEdit = NULL;
  psFunctyData->fNewColour = 0.0;
  psFunctyData->boPausedModal = FALSE;
  psFunctyData->boFileLoaded = FALSE;
  psFunctyData->szFilename = g_string_new (DEFAULT_FILENAME);
  psFunctyData->boFolderSet = FALSE;
  psFunctyData->szFolder = g_string_new ("");
  psFunctyData->boExportedModel = FALSE;
  psFunctyData->szExportModelName = g_string_new (DEFAULT_EXPORTMODELNAME);
  psFunctyData->boBinary = TRUE;
  psFunctyData->fExportMultiplier = 1.0f;
  psFunctyData->boExportScreen = TRUE;
  psFunctyData->boExportAlpha = TRUE;
  psFunctyData->eNextFunctionType = FUNCTYPE_INVALID;
  psFunctyData->boButtonBarHidden = FALSE;

  return psFunctyData;
}

void DeleteFunctyPersist (FunctyPersist * psFunctyData) {
  if (psFunctyData->psVisData) {
    DeleteVisPersist (psFunctyData->psVisData);
    psFunctyData->psVisData = NULL;
  }

  if (psFunctyData->szFilename) {
    g_string_free (psFunctyData->szFilename, TRUE);
  }
  if (psFunctyData->szFolder) {
    g_string_free (psFunctyData->szFolder, TRUE);
  }

  g_free (psFunctyData);
}

static void Realize (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GdkGLContext * psGlContext;
  GdkGLDrawable * psGlDrawable;
  bool boDrawSuccess;

  psGlContext = gtk_widget_get_gl_context (psWidget);
  psGlDrawable = gtk_widget_get_gl_drawable (psWidget);

  // Begin OpenGL setup
  boDrawSuccess = gdk_gl_drawable_gl_begin (psGlDrawable, psGlContext);
  if (boDrawSuccess) {
    Realise (psFunctyData->psVisData);

    gdk_gl_drawable_gl_end (psGlDrawable);
  }
  // End OpenGL setup
}

static gboolean ConfigureEvent (GtkWidget * psWidget, GdkEventConfigure * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GdkGLContext * psGlContext;
  GdkGLDrawable * psGlDrawable;
  GLfloat fWidth;
  GLfloat fHeight;
  bool boDrawSuccess;

  psGlContext = gtk_widget_get_gl_context (psWidget);
  psGlDrawable = gtk_widget_get_gl_drawable (psWidget);

  fWidth = psWidget->allocation.width;
  fHeight = psWidget->allocation.height;

  // Begin OpenGL window resize
  boDrawSuccess = gdk_gl_drawable_gl_begin (psGlDrawable, psGlContext);
  if (boDrawSuccess) {
    Reshape (fWidth, fHeight, psFunctyData->psVisData);

    gdk_gl_drawable_gl_end (psGlDrawable);
  }
  // End OpenGL window resize

  return boDrawSuccess;
}

static gboolean ExposeEvent (GtkWidget * psWidget, GdkEventExpose * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GdkGLContext * psGlContext;
  GdkGLDrawable * psGlDrawable;
  bool boDrawSuccess;
  bool boDoubleBuffered;

  psGlContext = gtk_widget_get_gl_context (psWidget);
  psGlDrawable = gtk_widget_get_gl_drawable (psWidget);

  // Begin OpenGL render
  boDrawSuccess = gdk_gl_drawable_gl_begin (psGlDrawable, psGlContext);
  if (boDrawSuccess) {
    // Redraw the visualisation
    Redraw (psFunctyData->psVisData);

    // Swap buffers
    boDoubleBuffered = gdk_gl_drawable_is_double_buffered (psGlDrawable);
    if (boDoubleBuffered) {
      gdk_gl_drawable_swap_buffers (psGlDrawable);
    }
    else {
      glFlush ();
    }

    gdk_gl_drawable_gl_end (psGlDrawable);
  }
  // End OpenGL render

  return boDrawSuccess;
}

static gboolean ButtonPressEvent (GtkWidget * psWidget, GdkEventButton * psEvent, gpointer * psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  //gboolean boMoving;

  //boMoving = GetMoving (psFunctyData->psVisData);

  Mouse (psEvent->button, psEvent->type, psEvent->x, psEvent->y, psFunctyData->psVisData);

  return FALSE;
}

static gboolean ButtonReleaseEvent (GtkWidget * psWidget, GdkEventButton * psEvent, gpointer * psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;

  Mouse (psEvent->button, psEvent->type, psEvent->x, psEvent->y, psFunctyData->psVisData);

  return FALSE;
}

/*
static gboolean MotionNotifyEvent (GtkWidget * psWidget, GdkEventMotion * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  //float fWidth;
  //float fHeight;
  float fX;
  float fY;

  //fWidth = psWidget->allocation.width;
  //fHeight = psWidget->allocation.height;
  fX = psEvent->x;
  fY = psEvent->y;

  Motion ((int)fX, (int)fY, psFunctyData->psVisData);
  gdk_window_invalidate_rect (psWidget->window, & psWidget->allocation, FALSE);

  return TRUE;
}
*/

static gboolean KeyPressEvent (GtkWidget * psWidget, GdkEventKey * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  guint uModifiers;

  uModifiers = gtk_accelerator_get_default_mod_mask ();

  switch (psEvent->keyval) {
		case 'b':
			ToggleButtonBar (psFunctyData);
			break;
    case GDK_Return:
      if ((psEvent->state & uModifiers) == GDK_MOD1_MASK) {
        ToggleFullScreenWindow (psFunctyData);
        SetViewWindowState (psFunctyData);
      }
      else {
        Key (psEvent->keyval, 0, 0, (psEvent->state & uModifiers), psFunctyData->psVisData);
        SetViewWindowState (psFunctyData);
      }
      break;
    default:
      Key (psEvent->keyval, 0, 0, (psEvent->state & uModifiers), psFunctyData->psVisData);
      SetViewWindowState (psFunctyData);
      break;
  }

  return TRUE;
}

static gboolean KeyReleaseEvent (GtkWidget * psWidget, GdkEventKey * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  guint uModifiers;

  uModifiers = gtk_accelerator_get_default_mod_mask ();

  KeyUp (psEvent->keyval, 0, 0, (psEvent->state & uModifiers), psFunctyData->psVisData);

  return TRUE;
}

static gboolean ScrollEvent (GtkWidget * psWidget, GdkEventScroll * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  float fViewRadius;
  float fLinkLen;
  float fCentring;
  float fRigidity;
  float fForce;
  float fResistance;
  guint uModifiers;

  uModifiers = gtk_accelerator_get_default_mod_mask ();

  if ((psEvent->state & uModifiers) == GDK_CONTROL_MASK) {
    switch (psEvent->direction) {
      case GDK_SCROLL_UP:
        ScaleFunctionRange (RANGE_SCALEUP, psFunctyData->psVisData);
        break;
      case GDK_SCROLL_DOWN:
        ScaleFunctionRange (RANGE_SCALEDOWN, psFunctyData->psVisData);
        break;
      default:
        // Do nothing
        break;
    }
  }
  else {
    GetDisplayProperties (& fViewRadius, & fLinkLen, & fCentring, & fRigidity, & fForce, & fResistance, psFunctyData->psVisData);

    switch (psEvent->direction) {
      case GDK_SCROLL_UP:
        fViewRadius += RADIUS_SCROLL;
        break;
      case GDK_SCROLL_DOWN:
        fViewRadius -= RADIUS_SCROLL;
        break;
      default:
        // Do nothing
        break;
    }

    SetDisplayProperties (fViewRadius, fLinkLen, fCentring, fRigidity, fForce, fResistance, 
      psFunctyData->psVisData);
  }

  return TRUE;
}

void ToggleButtonBar (FunctyPersist * psFunctyData) {
  GtkWidget * psWidget;

	psFunctyData->boButtonBarHidden = !psFunctyData->boButtonBarHidden;

	if (psFunctyData->boButtonBarHidden) {
		psWidget = glade_xml_get_widget (psFunctyData->psXML, "ButtonBar");
		gtk_widget_hide (psWidget);
	}
	else {
		psWidget = glade_xml_get_widget (psFunctyData->psXML, "ButtonBar");
		gtk_widget_show (psWidget);
	}
}

void ToggleFullScreenWindow (FunctyPersist * psFunctyData) {
  GtkWidget * psWindow;
  gboolean boCurrentState;

  boCurrentState = GetFullScreen (psFunctyData->psVisData);

  if (boCurrentState) {
    psWindow = glade_xml_get_widget (psFunctyData->psXML, "MainWindow");
    gtk_window_unfullscreen (GTK_WINDOW (psWindow));
  }
  else {
    psWindow = glade_xml_get_widget (psFunctyData->psXML, "MainWindow");
    gtk_window_fullscreen (GTK_WINDOW (psWindow));
  }

  ToggleFullScreen (psFunctyData->psVisData);
}

void SetFullScreenWindow (bool boFullScreen, FunctyPersist * psFunctyData) {
  gboolean boCurrentState;

  boCurrentState = GetFullScreen (psFunctyData->psVisData);

  if (boFullScreen != boCurrentState) {
    ToggleFullScreenWindow (psFunctyData);
  }
}

static gboolean Timeout (gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  int nX;
  int nY;
  GdkModifierType sState;

  if (psFunctyData->boTimeoutContinue) {
    // Update the pointer position
    // This differs from the standard motion callback, as it's called even if the pointer isn't moving
    // Update the range window
    if (UpdateRange (psFunctyData->psVisData)) {
      SetRangeWindowState (psFunctyData);
    }

    gdk_window_get_pointer (psFunctyData->psDrawingArea->window, & nX, & nY, & sState);
    Motion (nX, nY, psFunctyData->psVisData);

    // Call the idle function
    Idle (psFunctyData->psVisData);

    // Invalidate the drawing area
    gdk_window_invalidate_rect (psFunctyData->psDrawingArea->window, & psFunctyData->psDrawingArea->allocation, FALSE);

    // Update drawing area synchronously
    gdk_window_process_updates (psFunctyData->psDrawingArea->window, FALSE);
  }
  else {
    psFunctyData->TimeoutID = 0;
  }

  return psFunctyData->boTimeoutContinue;
}

static void TimeoutAdd (FunctyPersist * psFunctyData) {
  if (!psFunctyData->boPausedModal) {
    psFunctyData->boTimeoutContinue = TRUE;
    if (psFunctyData->TimeoutID == 0) {
      //psFunctyData->TimeoutID = g_timeout_add (UPDATE_TIMEOUT, Timeout, psFunctyData);
      psFunctyData->TimeoutID = g_timeout_add_full (G_PRIORITY_DEFAULT_IDLE, UPDATE_TIMEOUT, Timeout, psFunctyData, NULL);
      //psFunctyData->TimeoutID = g_timeout_add_full (G_PRIORITY_HIGH_IDLE, UPDATE_TIMEOUT, Timeout, psFunctyData, NULL);

      //psFunctyData->TimeoutID = g_idle_add (Timeout, psFunctyData);
    }
  }
}

static void TimeoutRemove (FunctyPersist * psFunctyData) {
  psFunctyData->boTimeoutContinue = FALSE;
}

static gboolean MapEvent (GtkWidget * psWidget, GdkEventAny * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;

  TimeoutAdd (psFunctyData);

  return TRUE;
}

static gboolean UnmapEvent (GtkWidget * psWidget, GdkEventAny * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;

  TimeoutRemove (psFunctyData);

  return TRUE;
}

static gboolean VisibilityNotifyEvent (GtkWidget * psWidget, GdkEventVisibility * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;

  if (psEvent->state == GDK_VISIBILITY_FULLY_OBSCURED) {
  	TimeoutRemove (psFunctyData);
  }
  else {
  	TimeoutAdd (psFunctyData);
  }

  return TRUE;
}

static gboolean ViewPress (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GtkWidget * psFunction;

  SetViewWindowState (psFunctyData);

  psFunction = glade_xml_get_widget (psFunctyData->psXML, "View");
  gtk_widget_show (psFunction);

  return TRUE;
}

static gboolean RangePress (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GtkWidget * psFunction;

  SetRangeWindowState (psFunctyData);

  psFunction = glade_xml_get_widget (psFunctyData->psXML, "Range");
  gtk_widget_show (psFunction);

  return TRUE;
}

static gboolean RangeSetPress (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GtkWidget * psFunction;

  psFunction = glade_xml_get_widget (psFunctyData->psXML, "Range");
  gtk_widget_hide (psFunction);

  return RangeSetApply (psWidget, psData);
}

static gboolean RangeSetApply (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  double afRange[6];

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XMin");
  afRange[0] = strtod (gtk_entry_get_text (GTK_ENTRY (psWidget)), NULL);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YMin");
  afRange[1] = strtod (gtk_entry_get_text (GTK_ENTRY (psWidget)), NULL);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZMin");
  afRange[2] = strtod (gtk_entry_get_text (GTK_ENTRY (psWidget)), NULL);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XWidth");
  afRange[3] = strtod (gtk_entry_get_text (GTK_ENTRY (psWidget)), NULL);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YWidth");
  afRange[4] = strtod (gtk_entry_get_text (GTK_ENTRY (psWidget)), NULL);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZWidth");
  afRange[5] = strtod (gtk_entry_get_text (GTK_ENTRY (psWidget)), NULL);

  SetVisRange (afRange, psFunctyData->psVisData);

  return TRUE;
}

gboolean RangeMinChange (GtkWidget *psWidget, GdkEventKey * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  double fMin;
  double fWidth;
  GtkWidget * psEntry;
  GString * szValue;

  szValue = g_string_new ("");

  psEntry = glade_xml_get_widget (psFunctyData->psXML, "XMin");
  fMin = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "XWidth");
  fWidth = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  g_string_printf (szValue, RANGE_FORMAT, (fMin + fWidth));
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "XMax");
  gtk_entry_set_text (GTK_ENTRY (psEntry), szValue->str);

  psEntry = glade_xml_get_widget (psFunctyData->psXML, "YMin");
  fMin = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "YWidth");
  fWidth = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  g_string_printf (szValue, RANGE_FORMAT, (fMin + fWidth));
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "YMax");
  gtk_entry_set_text (GTK_ENTRY (psEntry), szValue->str);

  psEntry = glade_xml_get_widget (psFunctyData->psXML, "ZMin");
  fMin = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "ZWidth");
  fWidth = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  g_string_printf (szValue, RANGE_FORMAT, (fMin + fWidth));
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "ZMax");
  gtk_entry_set_text (GTK_ENTRY (psEntry), szValue->str);

  g_string_free (szValue, TRUE);

  return FALSE;
}

gboolean RangeMaxChange (GtkWidget *psWidget, GdkEventKey * psEvent, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  double fMin;
  double fMax;
  GtkWidget * psEntry;
  GString * szValue;

  szValue = g_string_new ("");

  psEntry = glade_xml_get_widget (psFunctyData->psXML, "XMin");
  fMin = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "XMax");
  fMax = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  g_string_printf (szValue, RANGE_FORMAT, (fMax - fMin));
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "XWidth");
  gtk_entry_set_text (GTK_ENTRY (psEntry), szValue->str);

  psEntry = glade_xml_get_widget (psFunctyData->psXML, "YMin");
  fMin = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "YMax");
  fMax = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  g_string_printf (szValue, RANGE_FORMAT, (fMax - fMin));
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "YWidth");
  gtk_entry_set_text (GTK_ENTRY (psEntry), szValue->str);

  psEntry = glade_xml_get_widget (psFunctyData->psXML, "ZMin");
  fMin = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "ZMax");
  fMax = strtod (gtk_entry_get_text (GTK_ENTRY (psEntry)), NULL);
  g_string_printf (szValue, RANGE_FORMAT, (fMax - fMin));
  psEntry = glade_xml_get_widget (psFunctyData->psXML, "ZWidth");
  gtk_entry_set_text (GTK_ENTRY (psEntry), szValue->str);

  g_string_free (szValue, TRUE);

  return FALSE;
}
static gboolean PlotFunctionPress (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GtkWidget * psFunction;

  psFunction = glade_xml_get_widget (psFunctyData->psXML, "Function");
  gtk_widget_hide (psFunction);

  return PlotFunctionApply (psWidget, psData);
}

static gboolean PlotFunctionApply (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GtkWidget * psFunction;
  GtkWidget * psFunctionList;
  GtkWidget * psColourWidget;
  GtkWidget * psTexWidget;
  char const * szFunction;
  FuncPersist * psFuncData;
  GtkListStore * psFunctionListModel;
  GtkWidget * psWidgetAcc;
  double fAccuracy;
  char const * szRed;
  char const * szGreen;
  char const * szBlue;
  char const * szAlpha;
  char const * szXCentre;
  char const * szYCentre;
  char const * szZCentre;
  FUNCTYPE eFuncType;
  char const * szFilename;
  char const * szXScale;
  char const * szYScale;
  char const * szXOffset;
  char const * szYOffset;
  char const * szXFunction;
  char const * szYFunction;
  char const * szZFunction;

  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Red");
  szRed = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Green");
  szGreen = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Blue");
  szBlue = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Alpha");
  szAlpha = gtk_entry_get_text (GTK_ENTRY (psColourWidget));

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexFile");
  szFilename = gtk_entry_get_text (GTK_ENTRY (psTexWidget));
  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXScale");
  szXScale = gtk_entry_get_text (GTK_ENTRY (psTexWidget));
  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYScale");
  szYScale = gtk_entry_get_text (GTK_ENTRY (psTexWidget));
  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXOffset");
  szXOffset = gtk_entry_get_text (GTK_ENTRY (psTexWidget));
  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYOffset");
  szYOffset = gtk_entry_get_text (GTK_ENTRY (psTexWidget));

  psFunctionList = glade_xml_get_widget (psFunctyData->psXML, "FunctionList");

  if (psFunctyData->psFuncEdit) {
    eFuncType = GetFunctionType (psFunctyData->psFuncEdit);

    // Amend the selected function
    SetFunctionColours (szRed, szGreen, szBlue, szAlpha, psFunctyData->psFuncEdit);
		SetTextureValues (szFilename, szXScale, szYScale, szXOffset, szYOffset, psFunctyData->psFuncEdit);

    switch (eFuncType) {
    case FUNCTYPE_SPHERICAL:
		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "FunctionEdit");
		  szFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
	    SetFunction (szFunction, psFunctyData->psFuncEdit);

      // Set up the centre functions
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "XCentre");
      szXCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "YCentre");
      szYCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "ZCentre");
      szZCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));

      SphericalSetFunctionCentre (szXCentre, szYCentre, szZCentre, psFunctyData->psFuncEdit);

		  psWidgetAcc = glade_xml_get_widget (psFunctyData->psXML, "Accuracy");
		  fAccuracy = (1.0 / gtk_spin_button_get_value (GTK_SPIN_BUTTON (psWidgetAcc)));
		  SetFunctionAccuracy (fAccuracy, psFunctyData->psFuncEdit);
      break;
    case FUNCTYPE_CARTESIAN:
		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "FunctionEdit");
		  szFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
	    SetFunction (szFunction, psFunctyData->psFuncEdit);

		  psWidgetAcc = glade_xml_get_widget (psFunctyData->psXML, "Accuracy");
		  fAccuracy = (1.0 / gtk_spin_button_get_value (GTK_SPIN_BUTTON (psWidgetAcc)));
		  SetFunctionAccuracy (fAccuracy, psFunctyData->psFuncEdit);
    	break;
    case FUNCTYPE_CURVE:
		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "RadiusEdit");
		  szFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "CurveXEdit");
		  szXFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "CurveYEdit");
		  szYFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "CurveZEdit");
		  szZFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
			CurveSetFunction (szXFunction, szYFunction, szZFunction, szFunction, psFunctyData->psFuncEdit);

      // Set up the centre functions
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "XCentre");
      szXCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "YCentre");
      szYCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "ZCentre");
      szZCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));

      CurveSetFunctionCentre (szXCentre, szYCentre, szZCentre, psFunctyData->psFuncEdit);

	    psWidgetAcc = glade_xml_get_widget (psFunctyData->psXML, "AccuracyRadiusCurve");
		  fAccuracy = (1.0 / gtk_spin_button_get_value (GTK_SPIN_BUTTON (psWidgetAcc)));
			CurveSetFunctionAccuracyRadius (fAccuracy, psFunctyData->psFuncEdit);

		  psWidgetAcc = glade_xml_get_widget (psFunctyData->psXML, "AccuracyCurve");
		  fAccuracy = (1.0 / gtk_spin_button_get_value (GTK_SPIN_BUTTON (psWidgetAcc)));
		  SetFunctionAccuracy (fAccuracy, psFunctyData->psFuncEdit);
    	break;
    default:
      // Do nothing
      break;
    }

    PopulateVertices (psFunctyData->psFuncEdit);

    gdk_window_invalidate_rect (psFunctionList->window, NULL, TRUE);
  }
  else {
    // Add new function
    psFuncData = AddNewFunction (psFunctyData->eNextFunctionType, psFunctyData->psVisData);
    SetFunctionColours (szRed, szGreen, szBlue, szAlpha, psFuncData);

    switch (psFunctyData->eNextFunctionType) {
    case FUNCTYPE_SPHERICAL:
		  psWidgetAcc = glade_xml_get_widget (psFunctyData->psXML, "Accuracy");
		  fAccuracy = (1.0 / gtk_spin_button_get_value (GTK_SPIN_BUTTON (psWidgetAcc)));
			SetFunctionAccuracy (fAccuracy, psFuncData);

		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "FunctionEdit");
		  szFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
	    SetFunction (szFunction, psFuncData);

      // Set up the centre functions
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "XCentre");
      szXCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "YCentre");
      szYCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "ZCentre");
      szZCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));

      SphericalSetFunctionCentre (szXCentre, szYCentre, szZCentre, psFuncData);
      break;
    case FUNCTYPE_CARTESIAN:
		  psWidgetAcc = glade_xml_get_widget (psFunctyData->psXML, "Accuracy");
		  fAccuracy = (1.0 / gtk_spin_button_get_value (GTK_SPIN_BUTTON (psWidgetAcc)));
			SetFunctionAccuracy (fAccuracy, psFuncData);

		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "FunctionEdit");
		  szFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
	    SetFunction (szFunction, psFuncData);
    	break;
    case FUNCTYPE_CURVE:
	    psWidgetAcc = glade_xml_get_widget (psFunctyData->psXML, "AccuracyRadiusCurve");
		  fAccuracy = (1.0 / gtk_spin_button_get_value (GTK_SPIN_BUTTON (psWidgetAcc)));
			CurveSetFunctionAccuracyRadius (fAccuracy, psFuncData);

		  psWidgetAcc = glade_xml_get_widget (psFunctyData->psXML, "AccuracyCurve");
		  fAccuracy = (1.0 / gtk_spin_button_get_value (GTK_SPIN_BUTTON (psWidgetAcc)));
		  SetFunctionAccuracy (fAccuracy, psFuncData);

		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "RadiusEdit");
		  szFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "CurveXEdit");
		  szXFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "CurveYEdit");
		  szYFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
		  psFunction = glade_xml_get_widget (psFunctyData->psXML, "CurveZEdit");
		  szZFunction = gtk_entry_get_text (GTK_ENTRY (psFunction));
			CurveSetFunction (szXFunction, szYFunction, szZFunction, szFunction, psFuncData);

      // Set up the centre functions
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "XCentre");
      szXCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "YCentre");
      szYCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));
      psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "ZCentre");
      szZCentre = gtk_entry_get_text (GTK_ENTRY (psColourWidget));

      CurveSetFunctionCentre (szXCentre, szYCentre, szZCentre, psFuncData);
    	break;
    default:
      // Do nothing
      break;
    }

    TransferFunctionRange (psFuncData, psFunctyData->psVisData);

    psFunctionListModel = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (psFunctionList)));
    AddFunction (psFuncData, psFunctionListModel);

    psFunctyData->psFuncEdit = psFuncData;
  }

	FunctionShadersRegenerate (psFunctyData->psFuncEdit);

  SetColourButton (psFunctyData);

  return TRUE;
}

static void ClearAllFunctions (FunctyPersist * psFunctyData) {
  GtkWidget * psFunction;
  GtkListStore * psFunctionListModel;
  GtkWidget * psFunctionList;

  // Close the selected function window
  psFunction = glade_xml_get_widget (psFunctyData->psXML, "Function");
  gtk_widget_hide (psFunction);

  // No function selected
  psFunctyData->psFuncEdit = NULL;

  // Remove all of the rows from the function list
  psFunctionList = glade_xml_get_widget (psFunctyData->psXML, "FunctionList");
  psFunctionListModel = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (psFunctionList)));
  gtk_list_store_clear (psFunctionListModel);

  // Delete all of the functions
  DeleteAllFunctions (psFunctyData->psVisData);
}

static gboolean DeleteFunctionPress (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GtkWidget * psFunction;
  GtkWidget * psFunctionList;
  FuncPersist * psFuncData;
  GtkListStore * psFunctionListModel;
  GtkTreeSelection * psSelection;
  gboolean boFound;
  GtkTreeModel * psTreeModel;
  GtkTreeIter sTreeIter;

  psFunctionList = glade_xml_get_widget (psFunctyData->psXML, "FunctionList");

  psSelection = gtk_tree_view_get_selection  (GTK_TREE_VIEW (psFunctionList));

  boFound = gtk_tree_selection_get_selected (psSelection, & psTreeModel, & sTreeIter);

  if (boFound) {
    gtk_tree_model_get (psTreeModel, & sTreeIter, FUNCSCOL_FUNCDATA, & psFuncData, -1);

    // Only remove it if it is a function, and not the "New" option
    if (psFuncData) {
      if (psFuncData == psFunctyData->psFuncEdit) {
        // Close the selected function window
        psFunction = glade_xml_get_widget (psFunctyData->psXML, "Function");
        gtk_widget_hide (psFunction);

        // No function selected
        psFunctyData->psFuncEdit = NULL;
      }

      // Remove the currently selected row from the UI list
      psFunctionListModel = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (psFunctionList)));
      gtk_list_store_remove (psFunctionListModel, & sTreeIter);

      // Delete the function from the function list
      DeleteFunction (psFuncData, psFunctyData->psVisData);
    }
  }

  return TRUE;
}

static gboolean ToggleSpinPress (GtkToggleButton * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  bool boActive;

  boActive = gtk_toggle_button_get_active (psWidget);

  SetSpin (boActive, psFunctyData->psVisData);

  return TRUE;
}

static gboolean ToggleInvertPress (GtkToggleButton * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  bool boActive;

  boActive = gtk_toggle_button_get_active (psWidget);

  SetClearWhite (boActive, psFunctyData->psVisData);

  return TRUE;
}

static gboolean ToggleFullscreenPress (GtkToggleButton * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  bool boActive;

  boActive = gtk_toggle_button_get_active (psWidget);

  SetFullScreenWindow (boActive, psFunctyData);

  return TRUE;
}

static gboolean ToggleDrawAxesPress (GtkToggleButton * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  bool boActive;

  boActive = gtk_toggle_button_get_active (psWidget);

  SetDrawAxes (boActive, psFunctyData->psVisData);

  return TRUE;
}

static gboolean ToggleWireframePress (GtkToggleButton * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  bool boActive;

  boActive = gtk_toggle_button_get_active (psWidget);

  SetWireframe (boActive, psFunctyData->psVisData);

  return TRUE;
}

void SetViewWindowState (FunctyPersist * psFunctyData) {
  GtkWidget * psWidget;
  bool boActive;

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "DrawAxes");
  boActive = GetDrawAxes (psFunctyData->psVisData);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (psWidget), boActive);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Spin");
  boActive = GetSpin (psFunctyData->psVisData);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (psWidget), boActive);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Invert");
  boActive = GetClearWhite (psFunctyData->psVisData);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (psWidget), boActive);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Wireframe");
  boActive = GetWireframe (psFunctyData->psVisData);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (psWidget), boActive);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Fullscreen");
  boActive = GetFullScreen (psFunctyData->psVisData);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (psWidget), boActive);
}

void SetRangeWindowState (FunctyPersist * psFunctyData) {
  GtkWidget * psWidget;
  GString * szValue;
  double afRange[6];

  GetVisRange (afRange, psFunctyData->psVisData);

  szValue = g_string_new ("");

  g_string_printf (szValue, RANGE_FORMAT, afRange[0]);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XMin");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, RANGE_FORMAT, afRange[1]);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YMin");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, RANGE_FORMAT, afRange[2]);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZMin");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, RANGE_FORMAT, afRange[3]);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XWidth");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, RANGE_FORMAT, afRange[4]);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YWidth");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, RANGE_FORMAT, afRange[5]);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZWidth");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, RANGE_FORMAT, afRange[0] + afRange[3]);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XMax");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, RANGE_FORMAT, afRange[1] + afRange[4]);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YMax");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, RANGE_FORMAT, afRange[2] + afRange[5]);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZMax");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_free (szValue, TRUE);
}

void PopulateFunctionList (GtkListStore * psFunctionListModel, FunctyPersist * psFunctyData) {
  GSList * psFuncList;

  psFuncList = GetFunctionList (psFunctyData->psVisData);
  //AddFunction (NULL, psFunctionListModel);

  while (psFuncList) {
    AddFunction ((FuncPersist *)(psFuncList->data), psFunctionListModel);
    psFuncList = g_slist_next (psFuncList);
  }
}

void AddFunction (FuncPersist * psFuncData, GtkListStore * psFunctionListModel) {
  GtkTreeIter sListIter;

  gtk_list_store_append (psFunctionListModel, & sListIter);
  gtk_list_store_set (psFunctionListModel, & sListIter, FUNCSCOL_FUNCDATA, psFuncData, FUNCSCOL_FUNCTION, "", -1);
}

void FunctionCellDataFunc (GtkTreeViewColumn * psTreeColumn, GtkCellRenderer * psCell, GtkTreeModel * psTreeModel, GtkTreeIter * psIter, gpointer psData) {
  FuncPersist * psFuncData;
  char const * szFunction;
  gtk_tree_model_get (psTreeModel, psIter, FUNCSCOL_FUNCDATA, & psFuncData, -1);

  if (psFuncData) {
    szFunction = GetFunctionString (psFuncData);
    g_object_set (psCell, "text", szFunction, NULL);
  }
  else {
    g_object_set (psCell, "text", "New...", NULL);
  }
}

void EditFunctionRowDoubleClick (GtkTreeView * psTreeView, GtkTreePath * psPath, GtkTreeViewColumn * psColumn, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  gboolean boFound;
  GtkTreeModel * psTreeModel;
  GtkTreeIter sTreeIter;

  // Populate the Function window using information taken from the
  // row clicked on by the user

  psTreeModel = gtk_tree_view_get_model (psTreeView);

  boFound = gtk_tree_model_get_iter (psTreeModel, & sTreeIter, psPath);

  if (boFound) {
    EditFunction (psTreeModel, & sTreeIter, psFunctyData);
  }
}

static gboolean EditFunctionPress (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GtkWidget * psFunctionList;
  GtkTreeSelection * psSelection;
  gboolean boFound;
  GtkTreeModel * psTreeModel;
  GtkTreeIter sTreeIter;

  // Populate the Function window using information taken from the
  // currently selected function row

  psFunctionList = glade_xml_get_widget (psFunctyData->psXML, "FunctionList");

  psSelection = gtk_tree_view_get_selection  (GTK_TREE_VIEW (psFunctionList));

  boFound = gtk_tree_selection_get_selected (psSelection, & psTreeModel, & sTreeIter);

  if (boFound) {
    EditFunction (psTreeModel, & sTreeIter, psFunctyData);
  }

  return TRUE;
}

void EditFunction (GtkTreeModel * psTreeModel, GtkTreeIter * psTreeIter, FunctyPersist * psFunctyData) {
  GtkWidget * psFunctionEdit;
  FuncPersist * psFuncData;
  GtkWidget * psWidget;
  char const * szFunction;
  double fAccuracy;
  FUNCTYPE eFuncType;
  GtkWidget * psNotebook;
  int nPageRemoved;
  int nPage;

  // Populate the Function window using information taken from the supplied row

  gtk_tree_model_get (psTreeModel, psTreeIter, FUNCSCOL_FUNCDATA, & psFuncData, -1);
  psFunctyData->psFuncEdit = psFuncData;

  psFunctionEdit = glade_xml_get_widget (psFunctyData->psXML, "FunctionEdit");

	// Find the notebook to allow us to sort out the visible tabs
	psNotebook = glade_xml_get_widget (psFunctyData->psXML, "FunctionNotebook");

  if (psFuncData) {
    // Set the function edit box
    eFuncType = GetFunctionType (psFuncData);

    szFunction = GetFunctionString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psFunctionEdit), szFunction);

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "Red");
    szFunction = GetRedString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "Green");
    szFunction = GetGreenString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "Blue");
    szFunction = GetBlueString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "Alpha");
    szFunction = GetAlphaString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

    fAccuracy = GetFunctionAccuracy (psFuncData);
    psWidget = glade_xml_get_widget (psFunctyData->psXML, "Accuracy");
    gtk_spin_button_set_value (GTK_SPIN_BUTTON (psWidget), (1.0 / fAccuracy));

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexFile");
    szFunction = GetTexFileString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXScale");
    szFunction = GetTexXScaleString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYScale");
    szFunction = GetTexYScaleString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXOffset");
    szFunction = GetTexXOffsetString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYOffset");
    szFunction = GetTexYOffsetString (psFunctyData->psFuncEdit);
    gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

    SetColourButton (psFunctyData);

    switch (eFuncType) {
    case FUNCTYPE_CARTESIAN:
			// Hide the centre tab
			psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCentre");
			gtk_widget_hide (psWidget);

			// Show the function tab
			psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsFunction");
			gtk_widget_show (psWidget);

			// Hide the curve tab
			psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCurve");
			nPageRemoved = gtk_notebook_page_num (GTK_NOTEBOOK (psNotebook), psWidget);
			nPage = gtk_notebook_get_current_page (GTK_NOTEBOOK(psNotebook));
			if (nPage == nPageRemoved) {
				gtk_notebook_set_current_page (GTK_NOTEBOOK(psNotebook), (nPage - 1));
			}
			gtk_widget_hide (psWidget);

      // Set the window title
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
      gtk_window_set_title (GTK_WINDOW (psWidget), "Edit Cartesian function");
      break;
    case FUNCTYPE_SPHERICAL:
			// Show the centre tab
			psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCentre");
			gtk_widget_show (psWidget);

			// Show the function tab
			psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsFunction");
			gtk_widget_show (psWidget);

			// Hide the curve tab
			psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCurve");
			nPageRemoved = gtk_notebook_page_num (GTK_NOTEBOOK (psNotebook), psWidget);
			nPage = gtk_notebook_get_current_page (GTK_NOTEBOOK(psNotebook));
			if (nPage == nPageRemoved) {
				gtk_notebook_set_current_page (GTK_NOTEBOOK(psNotebook), (nPage - 1));
			}
			gtk_widget_hide (psWidget);

      // Set the window title
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
      gtk_window_set_title (GTK_WINDOW (psWidget), "Edit Spherical function");

      // Set up the centre values
      szFunction = SphericalGetXCentreString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "XCentre");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

      szFunction = SphericalGetYCentreString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "YCentre");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

      szFunction = SphericalGetZCentreString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZCentre");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);
      break;
    case FUNCTYPE_CURVE:
			// Show the centre tab
			psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCentre");
			gtk_widget_show (psWidget);

			// Show the curve tab
			psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCurve");
			gtk_widget_show (psWidget);

			// Hide the function tab
			psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsFunction");
			nPageRemoved = gtk_notebook_page_num (GTK_NOTEBOOK (psNotebook), psWidget);
			nPage = gtk_notebook_get_current_page (GTK_NOTEBOOK(psNotebook));
			if (nPage == nPageRemoved) {
				gtk_notebook_set_current_page (GTK_NOTEBOOK(psNotebook), (nPage + 1));
			}
			gtk_widget_hide (psWidget);

      // Set the window title
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
      gtk_window_set_title (GTK_WINDOW (psWidget), "Edit Curve function");

      // Set up the curve values
      szFunction = CurveGetRadiusString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "RadiusEdit");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

      szFunction = CurveGetXFunctionString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "CurveXEdit");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

      szFunction = CurveGetYFunctionString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "CurveYEdit");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

      szFunction = CurveGetZFunctionString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "CurveZEdit");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

		  fAccuracy = GetFunctionAccuracy (psFuncData);
		  psWidget = glade_xml_get_widget (psFunctyData->psXML, "AccuracyCurve");
		  gtk_spin_button_set_value (GTK_SPIN_BUTTON (psWidget), (1.0 / fAccuracy));

		  fAccuracy = CurveGetFunctionAccuracyRadius (psFuncData);
		  psWidget = glade_xml_get_widget (psFunctyData->psXML, "AccuracyRadiusCurve");
		  gtk_spin_button_set_value (GTK_SPIN_BUTTON (psWidget), (1.0 / fAccuracy));

      // Set up the centre values
      szFunction = CurveGetXCentreString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "XCentre");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

      szFunction = CurveGetYCentreString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "YCentre");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);

      szFunction = CurveGetZCentreString (psFunctyData->psFuncEdit);
      psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZCentre");
      gtk_entry_set_text (GTK_ENTRY (psWidget), szFunction);
    	break;
    default:
      // Do nothing
      break;
    }

    psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
    gtk_widget_show (psWidget);
  }
  else {
    // If the function data is NULL we'll start to edit a new function
    // This should never happen the way the interface is currently set up
    NewCartesianFunction (psFunctyData);
    // Set the window title
    psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
    gtk_window_set_title (GTK_WINDOW (psWidget), "Edit Cartesian function");

		// Hide the centre tab
		psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCentre");
		gtk_widget_hide (psWidget);

		// Show the function tab
		psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsFunction");
		gtk_widget_show (psWidget);

		// Hide the curve tab
		psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCurve");
		nPageRemoved = gtk_notebook_page_num (GTK_NOTEBOOK (psNotebook), psWidget);
		nPage = gtk_notebook_get_current_page (GTK_NOTEBOOK(psNotebook));
		if (nPage == nPageRemoved) {
			gtk_notebook_set_current_page (GTK_NOTEBOOK(psNotebook), (nPage - 1));
		}
		gtk_widget_hide (psWidget);
  }
}

void NewCartesianFunction (FunctyPersist * psFunctyData) {
  GtkWidget * psWidget;
  GtkWidget * psFunctionEdit;
  GtkWidget * psColourWidget;
  GtkWidget * psTexWidget;
  GtkWidget * psNotebook;
  GString * szValue;
  GdkColor sColour;
  double fRed;
  double fGreen;
  double fBlue;
  double fAlpha;
  int nPageRemoved;
  int nPage;

  psFunctyData->eNextFunctionType = FUNCTYPE_CARTESIAN;
  psFunctyData->psFuncEdit = NULL;

  psFunctionEdit = glade_xml_get_widget (psFunctyData->psXML, "FunctionEdit");

  gtk_entry_set_text (GTK_ENTRY (psFunctionEdit), FUNCTION_NEW_CARTESIAN);

  szValue = g_string_new ("");

  GetNextColour (& fRed, & fGreen, & fBlue, & fAlpha, psFunctyData);

  sColour.red = (guint16)floor ((65535.0 * fRed) + 0.5);
  g_string_printf (szValue, "%.3f", fRed);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Red");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  sColour.green = (guint16)floor ((65535.0 * fGreen) + 0.5);
  g_string_printf (szValue, "%.3f", fGreen);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Green");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  sColour.blue = (guint16)floor ((65535.0 * fBlue) + 0.5);
  g_string_printf (szValue, "%.3f", fBlue);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Blue");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  g_string_printf (szValue, "%.3f", fAlpha);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Alpha");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  g_string_free (szValue, TRUE);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Colour");
  gtk_color_button_set_color (GTK_COLOR_BUTTON (psWidget), & sColour);
  gtk_color_button_set_alpha (GTK_COLOR_BUTTON (psWidget), (guint16)floor ((65535.0 * ALPHA_NEW) + 0.5));

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexFile");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXScale");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "1.0");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYScale");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "1.0");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXOffset");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "0.0");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYOffset");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "0.0");

	// Sort out the visible tabs of the notebook
	psNotebook = glade_xml_get_widget (psFunctyData->psXML, "FunctionNotebook");

  // Hide the centre tab
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCentre");
  gtk_widget_hide (psWidget);

  // Show the function tab
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsFunction");
  gtk_widget_show (psWidget);

  // Hide the curve tab
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCurve");
  nPageRemoved = gtk_notebook_page_num (GTK_NOTEBOOK (psNotebook), psWidget);
  nPage = gtk_notebook_get_current_page (GTK_NOTEBOOK(psNotebook));
  if (nPage == nPageRemoved) {
	  gtk_notebook_set_current_page (GTK_NOTEBOOK(psNotebook), (nPage - 1));
	}
  gtk_widget_hide (psWidget);

  // Set the window title
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
  gtk_window_set_title (GTK_WINDOW (psWidget), "New Cartesian function");

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
  gtk_widget_show (psWidget);
}

static gboolean AddCartesianPress (GtkColorButton * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;

  NewCartesianFunction (psFunctyData);

  return TRUE;
}

void NewSphericalFunction (FunctyPersist * psFunctyData) {
  GtkWidget * psWidget;
  GtkWidget * psFunctionEdit;
  GtkWidget * psColourWidget;
  GtkWidget * psTexWidget;
  GtkWidget * psNotebook;
  GString * szValue;
  GdkColor sColour;
  double fRed;
  double fGreen;
  double fBlue;
  double fAlpha;
  int nPageRemoved;
  int nPage;

  psFunctyData->eNextFunctionType = FUNCTYPE_SPHERICAL;
  psFunctyData->psFuncEdit = NULL;

  psFunctionEdit = glade_xml_get_widget (psFunctyData->psXML, "FunctionEdit");

  gtk_entry_set_text (GTK_ENTRY (psFunctionEdit), FUNCTION_NEW_SPHERICAL);

  szValue = g_string_new ("");

  GetNextColour (& fRed, & fGreen, & fBlue, & fAlpha, psFunctyData);

  sColour.red = (guint16)floor ((65535.0 * fRed) + 0.5);
  g_string_printf (szValue, "%.3f", fRed);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Red");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  sColour.green = (guint16)floor ((65535.0 * fGreen) + 0.5);
  g_string_printf (szValue, "%.3f", fGreen);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Green");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  sColour.blue = (guint16)floor ((65535.0 * fBlue) + 0.5);
  g_string_printf (szValue, "%.3f", fBlue);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Blue");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  g_string_printf (szValue, "%.3f", fAlpha);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Alpha");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Colour");
  gtk_color_button_set_color (GTK_COLOR_BUTTON (psWidget), & sColour);
  gtk_color_button_set_alpha (GTK_COLOR_BUTTON (psWidget), (guint16)floor ((65535.0 * ALPHA_NEW) + 0.5));

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexFile");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXScale");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "1.0");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYScale");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "1.0");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXOffset");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "0.0");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYOffset");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "0.0");

  // Set up the centre values
  g_string_printf (szValue, "%.3f", SPHERICAL_XCENTRE);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XCentre");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, "%.3f", SPHERICAL_YCENTRE);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YCentre");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, "%.3f", SPHERICAL_ZCENTRE);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZCentre");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

	// Sort out the visible tabs of the notebook
	psNotebook = glade_xml_get_widget (psFunctyData->psXML, "FunctionNotebook");

  // Show the centre tab
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCentre");
  gtk_widget_show (psWidget);

  // Show the function tab
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsFunction");
  gtk_widget_show (psWidget);

  // Hide the curve tab
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCurve");
  nPageRemoved = gtk_notebook_page_num (GTK_NOTEBOOK (psNotebook), psWidget);
  nPage = gtk_notebook_get_current_page (GTK_NOTEBOOK(psNotebook));
  if (nPage == nPageRemoved) {
	  gtk_notebook_set_current_page (GTK_NOTEBOOK(psNotebook), (nPage - 1));
	}
  gtk_widget_hide (psWidget);

  // Set the window title
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
  gtk_window_set_title (GTK_WINDOW (psWidget), "New Spherical function");

  g_string_free (szValue, TRUE);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
  gtk_widget_show (psWidget);
}

static gboolean AddSphericalPress (GtkColorButton * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;

  NewSphericalFunction (psFunctyData);

  return TRUE;
}

void NewCurveFunction (FunctyPersist * psFunctyData) {
  GtkWidget * psWidget;
  GtkWidget * psRadiusEdit;
  GtkWidget * psCurveXEdit;
  GtkWidget * psCurveYEdit;
  GtkWidget * psCurveZEdit;
  GtkWidget * psColourWidget;
  GtkWidget * psTexWidget;
  GtkWidget * psNotebook;
  GString * szValue;
  GdkColor sColour;
  double fRed;
  double fGreen;
  double fBlue;
  double fAlpha;
  int nPageRemoved;
  int nPage;

  psFunctyData->eNextFunctionType = FUNCTYPE_CURVE;
  psFunctyData->psFuncEdit = NULL;

  psRadiusEdit = glade_xml_get_widget (psFunctyData->psXML, "RadiusEdit");
  psCurveXEdit = glade_xml_get_widget (psFunctyData->psXML, "CurveXEdit");
  psCurveYEdit = glade_xml_get_widget (psFunctyData->psXML, "CurveYEdit");
  psCurveZEdit = glade_xml_get_widget (psFunctyData->psXML, "CurveZEdit");

  gtk_entry_set_text (GTK_ENTRY (psRadiusEdit), FUNCTION_NEW_CURVE_RADIUS);
  gtk_entry_set_text (GTK_ENTRY (psCurveXEdit), FUNCTION_NEW_CURVE_X);
  gtk_entry_set_text (GTK_ENTRY (psCurveYEdit), FUNCTION_NEW_CURVE_Y);
  gtk_entry_set_text (GTK_ENTRY (psCurveZEdit), FUNCTION_NEW_CURVE_Z);

  szValue = g_string_new ("");

  GetNextColour (& fRed, & fGreen, & fBlue, & fAlpha, psFunctyData);

  sColour.red = (guint16)floor ((65535.0 * fRed) + 0.5);
  g_string_printf (szValue, "%.3f", fRed);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Red");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  sColour.green = (guint16)floor ((65535.0 * fGreen) + 0.5);
  g_string_printf (szValue, "%.3f", fGreen);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Green");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  sColour.blue = (guint16)floor ((65535.0 * fBlue) + 0.5);
  g_string_printf (szValue, "%.3f", fBlue);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Blue");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  g_string_printf (szValue, "%.3f", fAlpha);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Alpha");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Colour");
  gtk_color_button_set_color (GTK_COLOR_BUTTON (psWidget), & sColour);
  gtk_color_button_set_alpha (GTK_COLOR_BUTTON (psWidget), (guint16)floor ((65535.0 * ALPHA_NEW) + 0.5));

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexFile");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXScale");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "1.0");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYScale");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "1.0");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXOffset");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "0.0");

  psTexWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYOffset");
  gtk_entry_set_text (GTK_ENTRY (psTexWidget), "0.0");

  // Set up the centre values
  g_string_printf (szValue, "%.3f", CURVE_XCENTRE);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XCentre");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, "%.3f", CURVE_YCENTRE);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YCentre");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

  g_string_printf (szValue, "%.3f", CURVE_ZCENTRE);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZCentre");
  gtk_entry_set_text (GTK_ENTRY (psWidget), szValue->str);

	// Sort out the visible tabs of the notebook
	psNotebook = glade_xml_get_widget (psFunctyData->psXML, "FunctionNotebook");

  // Show the centre tab
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCentre");
  gtk_widget_show (psWidget);

  // Show the curve tab
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsCurve");
  gtk_widget_show (psWidget);

  // Hide the function tab
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TabContentsFunction");
  nPageRemoved = gtk_notebook_page_num (GTK_NOTEBOOK (psNotebook), psWidget);
  nPage = gtk_notebook_get_current_page (GTK_NOTEBOOK(psNotebook));
  if (nPage == nPageRemoved) {
	  gtk_notebook_set_current_page (GTK_NOTEBOOK(psNotebook), (nPage + 1));
	}
  gtk_widget_hide (psWidget);

  // Set the window title
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
  gtk_window_set_title (GTK_WINDOW (psWidget), "New Curve function");

  g_string_free (szValue, TRUE);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Function");
  gtk_widget_show (psWidget);
}

static gboolean AddCurvePress (GtkColorButton * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;

  NewCurveFunction (psFunctyData);

  return TRUE;
}

void GetNextColour (double * pfRed, double * pfGreen, double * pfBlue, double * pfAlpha, FunctyPersist * psFunctyData) {
  if (pfRed) {
    * pfRed = 0.4 + 0.6 * ((cos (psFunctyData->fNewColour) + 1.0) / 2.0);
  }
  if (pfGreen) {
    * pfGreen = 0.4 + 0.6 * ((cos (psFunctyData->fNewColour + (2.0 * M_PI / 3.0)) + 1.0) / 2.0);
  }
  if (pfBlue) {
    * pfBlue = 0.4 + 0.6 * ((cos (psFunctyData->fNewColour + (4.0 * M_PI / 3.0)) + 1.0) / 2.0);
  }
  if (pfAlpha) {
    * pfAlpha = ALPHA_NEW;
  }

  psFunctyData->fNewColour += COLOUR_NEW_SHIFT;
}

void SetColourButton (FunctyPersist * psFunctyData) {
  GtkWidget * psWidget;
  bool boColour;
  GLfloat afGraphColour[4];
  GdkColor sColour;

  boColour = GetColour (afGraphColour, psFunctyData->psFuncEdit);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Colour");
  if (boColour) {
    sColour.red = (guint16)floor ((65535.0 * afGraphColour[0]) + 0.5);
    sColour.green = (guint16)floor ((65535.0 * afGraphColour[1]) + 0.5);
    sColour.blue = (guint16)floor ((65535.0 * afGraphColour[2]) + 0.5);
    gtk_color_button_set_color (GTK_COLOR_BUTTON (psWidget), & sColour);

    gtk_color_button_set_alpha (GTK_COLOR_BUTTON (psWidget), (guint16)floor ((65535.0 * afGraphColour[3]) + 0.5));
  }
  else {
    gtk_color_button_set_alpha (GTK_COLOR_BUTTON (psWidget), (guint16)0);
  }
}

void SetColourButtonCallback (GtkColorButton * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GdkColor sColour;
  guint16 uAlpha;
  GtkWidget * psColourWidget;
  GString * szValue;

  szValue = g_string_new ("");

  gtk_color_button_get_color (psWidget, & sColour);
  uAlpha = gtk_color_button_get_alpha (psWidget);

  g_string_printf (szValue, "%.3f", (double)sColour.red / 65535.0);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Red");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  g_string_printf (szValue, "%.3f", (double)sColour.green / 65535.0);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Green");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  g_string_printf (szValue, "%.3f", (double)sColour.blue / 65535.0);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Blue");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  g_string_printf (szValue, "%.3f", (double)uAlpha / 65535.0);
  psColourWidget = glade_xml_get_widget (psFunctyData->psXML, "Alpha");
  gtk_entry_set_text (GTK_ENTRY (psColourWidget), szValue->str);

  g_string_free (szValue, TRUE);
}

static gboolean LoadFilePress (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GtkWidget * psDialogue;
  GtkWindow * psParent;
  char * szFilename;
  char * szFolder;
  GtkListStore * psFunctionListModel;
  GtkWidget * psFunctionList;
  bool boLoaded;
  GtkFileFilter * psFilterXML;
  GtkFileFilter * psFilterAll;

  psFilterXML = gtk_file_filter_new ();
  gtk_file_filter_add_mime_type (psFilterXML, "text/xml");
  gtk_file_filter_add_mime_type (psFilterXML, "application/xml");
  gtk_file_filter_set_name (psFilterXML, "XML Files");

  psFilterAll = gtk_file_filter_new ();
  gtk_file_filter_add_pattern (psFilterAll, "*");
  gtk_file_filter_set_name (psFilterAll, "All Files");


  psFunctionList = glade_xml_get_widget (psFunctyData->psXML, "FunctionList");
  psFunctionListModel = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (psFunctionList)));
  psParent = GTK_WINDOW (glade_xml_get_widget (psFunctyData->psXML, "MainWindow"));

  psDialogue = gtk_file_chooser_dialog_new ("Open File", psParent, 
    GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
    GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

  g_object_set (G_OBJECT (psDialogue), "local-only", FALSE, NULL);
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (psDialogue), psFilterXML);
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (psDialogue), psFilterAll);

  if (psFunctyData->boFolderSet) {
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (psDialogue), psFunctyData->szFolder->str);
  }

  PauseAnimationModal (TRUE, psFunctyData);

  if (gtk_dialog_run (GTK_DIALOG (psDialogue)) == GTK_RESPONSE_ACCEPT) {
    szFilename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (psDialogue));
    ClearAllFunctions (psFunctyData);
    boLoaded = LoadFile (szFilename, FALSE, psFunctyData);
    PopulateFunctionList (psFunctionListModel, psFunctyData);
    if (boLoaded) {
      g_string_assign (psFunctyData->szFilename, szFilename);

      szFolder = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (psDialogue));
      psFunctyData->boFolderSet = (szFolder != NULL);
      if (szFolder) {
        g_string_assign (psFunctyData->szFolder, szFolder);
        g_free (szFolder);
        szFolder = NULL;
      }

      // Set the window title using the file display name
      SetMainWindowTitle (szFilename, psFunctyData);
    }
    psFunctyData->boFileLoaded = boLoaded;
    g_free (szFilename);
    szFilename = NULL;
  }
  gtk_widget_destroy (psDialogue);

  PauseAnimationModal (FALSE, psFunctyData);

  return TRUE;
}

static gboolean SaveFilePress (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;
  GtkWidget * psDialogue;
  GtkWindow * psParent;
  char * szFilename;
  char * szFolder;
  bool boLoaded;
  GtkFileFilter * psFilterXML;
  GtkFileFilter * psFilterAll;

  psFilterXML = gtk_file_filter_new ();
  gtk_file_filter_add_mime_type (psFilterXML, "text/xml");
  gtk_file_filter_add_mime_type (psFilterXML, "application/xml");
  gtk_file_filter_set_name (psFilterXML, "XML Files");

  psFilterAll = gtk_file_filter_new ();
  gtk_file_filter_add_pattern (psFilterAll, "*");
  gtk_file_filter_set_name (psFilterAll, "All Files");

  psParent = GTK_WINDOW (glade_xml_get_widget (psFunctyData->psXML, "MainWindow"));

  psDialogue = gtk_file_chooser_dialog_new ("Save File", psParent, 
    GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
    GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);

  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (psDialogue), TRUE);
  g_object_set (G_OBJECT (psDialogue), "local-only", FALSE, NULL);
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (psDialogue), psFilterXML);
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (psDialogue), psFilterAll);

  PauseAnimationModal (TRUE, psFunctyData);

  if (psFunctyData->boFolderSet) {
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (psDialogue), psFunctyData->szFolder->str);
  }
  else {
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (psDialogue), g_get_home_dir ());
  }

  if (!psFunctyData->boFileLoaded) {
    gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (psDialogue), DEFAULT_FILENAME);
  }
  else {
    gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (psDialogue), psFunctyData->szFilename->str);
  }

  if (gtk_dialog_run (GTK_DIALOG (psDialogue)) == GTK_RESPONSE_ACCEPT) {
    szFilename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (psDialogue));
    boLoaded = SaveFile (szFilename, FALSE, psFunctyData);
    if (boLoaded) {
      g_string_assign (psFunctyData->szFilename, szFilename);

      szFolder = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (psDialogue));
      psFunctyData->boFolderSet = (szFolder != NULL);
      if (szFolder) {
        g_string_assign (psFunctyData->szFolder, szFolder);
        g_free (szFolder);
        szFolder = NULL;
      }

      // Set the window title using the file display name
      SetMainWindowTitle (szFilename, psFunctyData);
    }
    psFunctyData->boFileLoaded = boLoaded;
    g_free (szFilename);
    szFilename = NULL;
  }
  gtk_widget_destroy (psDialogue);

  PauseAnimationModal (FALSE, psFunctyData);

  return TRUE;
}

static gboolean ExportFilePress (GtkWidget * psWidget, gpointer psData) {
  FunctyPersist * psFunctyData = (FunctyPersist * )psData;

  GtkWidget * psDialogue;
  GtkWindow * psParent;
  char * szFilename;
  char * szFolder;
  bool boExported;
  GtkFileFilter * psFilterXML;
  GtkFileFilter * psFilterAll;
  GtkWidget * psBinary;
  GtkWidget * psMultiplier;
  GtkWidget * psScreenCoords;
  GtkWidget * psOptions;
  GtkWidget * psAlpha;
  GtkWidget * psWidgetAdd;
  GtkAdjustment * psAdjustment;

  psFilterXML = gtk_file_filter_new ();
  gtk_file_filter_add_pattern (psFilterXML, "*.ply");
  gtk_file_filter_add_mime_type (psFilterXML, "text/ply");
  gtk_file_filter_add_mime_type (psFilterXML, "application/ply");
  gtk_file_filter_set_name (psFilterXML, "Stanford Triangle Format");

  psFilterAll = gtk_file_filter_new ();
  gtk_file_filter_add_pattern (psFilterAll, "*");
  gtk_file_filter_set_name (psFilterAll, "All Files");

  psParent = GTK_WINDOW (glade_xml_get_widget (psFunctyData->psXML, "MainWindow"));

  psDialogue = gtk_file_chooser_dialog_new ("Export Model", psParent, 
    GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
    GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);

  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (psDialogue), TRUE);
  g_object_set (G_OBJECT (psDialogue), "local-only", FALSE, NULL);
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (psDialogue), psFilterXML);
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (psDialogue), psFilterAll);

  PauseAnimationModal (TRUE, psFunctyData);

  if (psFunctyData->boFolderSet) {
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (psDialogue), psFunctyData->szFolder->str);
  }
  else {
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (psDialogue), g_get_home_dir ());
  }

  if (!psFunctyData->boExportedModel) {
    gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (psDialogue), DEFAULT_EXPORTMODELNAME);
  }
  else {
    gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (psDialogue), psFunctyData->szExportModelName->str);
  }

	psOptions = gtk_hbox_new (FALSE, 8);
	gtk_widget_show (psOptions);

	psWidgetAdd = gtk_label_new ("Accuracy multiplier");
	gtk_label_set_justify (GTK_LABEL (psWidgetAdd), GTK_JUSTIFY_RIGHT);
	gtk_widget_show (psWidgetAdd);
	gtk_box_pack_start (GTK_BOX (psOptions), psWidgetAdd, FALSE, TRUE, 0);

	psAdjustment = GTK_ADJUSTMENT (gtk_adjustment_new (psFunctyData->fExportMultiplier, 0.00, 1000.0, 0.1, 1.0, 0.0));
	psMultiplier = gtk_spin_button_new (psAdjustment, 1.0, 3);
	gtk_widget_show (psMultiplier);
	gtk_box_pack_start (GTK_BOX (psOptions), psMultiplier, FALSE, TRUE, 0);

	psBinary = gtk_check_button_new_with_label ("Binary format");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (psBinary), psFunctyData->boBinary);
	gtk_widget_show (psBinary);
	gtk_box_pack_start (GTK_BOX (psOptions), psBinary, FALSE, TRUE, 0);

	psAlpha = gtk_check_button_new_with_label ("Include alpha");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (psAlpha), psFunctyData->boExportAlpha);
	gtk_widget_show (psAlpha);
	gtk_box_pack_start (GTK_BOX (psOptions), psAlpha, FALSE, TRUE, 0);

	psScreenCoords = gtk_check_button_new_with_label ("Screen coordinates");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (psScreenCoords), psFunctyData->boExportScreen);
	gtk_widget_show (psScreenCoords);
	gtk_box_pack_start (GTK_BOX (psOptions), psScreenCoords, FALSE, TRUE, 0);

	gtk_file_chooser_set_extra_widget (GTK_FILE_CHOOSER (psDialogue), psOptions);

  if (gtk_dialog_run (GTK_DIALOG (psDialogue)) == GTK_RESPONSE_ACCEPT) {
  	psFunctyData->boBinary = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (psBinary));
		psFunctyData->fExportMultiplier = gtk_spin_button_get_value (GTK_SPIN_BUTTON (psMultiplier));
		psFunctyData->boExportScreen = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (psScreenCoords));
		psFunctyData->boExportAlpha = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (psAlpha));
    szFilename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (psDialogue));
    boExported = ExportModelFile (szFilename, psFunctyData->boBinary, psFunctyData->boExportScreen, psFunctyData->boExportAlpha, psFunctyData->fExportMultiplier, psFunctyData->psVisData);
    if (boExported) {
      g_string_assign (psFunctyData->szExportModelName, szFilename);

      szFolder = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (psDialogue));
      psFunctyData->boFolderSet = (szFolder != NULL);
      if (szFolder) {
        g_string_assign (psFunctyData->szFolder, szFolder);
        g_free (szFolder);
        szFolder = NULL;
      }
    }
    psFunctyData->boExportedModel = boExported;
    g_free (szFilename);
    szFilename = NULL;
  }
  gtk_widget_destroy (psDialogue);

  PauseAnimationModal (FALSE, psFunctyData);

  return TRUE;
}

void PauseAnimationModal (bool boPauseModal, FunctyPersist * psFunctyData) {
  psFunctyData->boPausedModal = boPauseModal;
	if (boPauseModal) {
	  TimeoutRemove (psFunctyData);
	}
	else {
	  TimeoutAdd (psFunctyData);
	}
}


int main (int argc, char *argv[]) {
  GdkGLConfig * GlConfig;
  GtkWidget * psWindow;
  GtkWidget * psWidget;
  VisPersist * psVisData;
  FunctyPersist * psFunctyData;
  GtkListStore * psFunctionListModel;
  GtkCellRenderer * psListCellRenderer;
  GtkTreeViewColumn * psListColumn;

  // Initialise various libraries
  gtk_init (&argc, &argv);
  gtk_gl_init (&argc, &argv);
  glutInit (&argc, argv);

  // Create new persistent structures
  psVisData = NewVisPersist ();
  psFunctyData = NewFunctyPersist ();
  psFunctyData->psVisData = psVisData;

  // First try double buffering
  GlConfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH  | GDK_GL_MODE_DOUBLE | GDK_GL_MODE_ALPHA | GDK_GL_MODE_STENCIL);
  if (GlConfig == NULL) {
    g_print ("Failed to initialise double buffered visual; trying single buffer.\n");

    // If that fails, we'll try single buffered
    GlConfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH | GDK_GL_MODE_ALPHA | GDK_GL_MODE_STENCIL);
    if (GlConfig == NULL) {
  	  g_print ("OpenGL visual could not be initialised.\n");
  	  exit (1);
  	}
  }

  // Load the glade interface
  psFunctyData->psXML = glade_xml_new (FUNCTYDIR "/Functy.glade", NULL, NULL);

  // Main window
  psWindow = glade_xml_get_widget (psFunctyData->psXML, "MainWindow");
  gtk_container_set_reallocate_redraws (GTK_CONTAINER (psWindow), TRUE);
  psFunctyData->psDrawingArea = glade_xml_get_widget (psFunctyData->psXML, "DrawingArea");

  // Automatically redraw the window children change allocation
  gtk_container_set_reallocate_redraws (GTK_CONTAINER (psWindow), TRUE);

  // Set up OpenGL for the drawing area widget
  gtk_widget_set_gl_capability (psFunctyData->psDrawingArea, GlConfig, NULL, TRUE, GDK_GL_RGBA_TYPE);

  // Connect up relevant signals
  glade_xml_signal_autoconnect (psFunctyData->psXML);

  g_signal_connect_after (G_OBJECT (psFunctyData->psDrawingArea), "realize", G_CALLBACK (Realize), 
    (gpointer)psFunctyData);
  g_signal_connect (G_OBJECT (psFunctyData->psDrawingArea), "configure_event", G_CALLBACK (ConfigureEvent), 
    (gpointer)psFunctyData);
  g_signal_connect (G_OBJECT (psFunctyData->psDrawingArea), "expose_event", G_CALLBACK (ExposeEvent), 
    (gpointer)psFunctyData);

  g_signal_connect (G_OBJECT (psFunctyData->psDrawingArea), "button_press_event", 
    G_CALLBACK (ButtonPressEvent), (gpointer)psFunctyData);
  g_signal_connect (G_OBJECT (psFunctyData->psDrawingArea), "button_release_event", 
    G_CALLBACK (ButtonReleaseEvent), (gpointer)psFunctyData);
  //g_signal_connect (G_OBJECT (psFunctyData->psDrawingArea), "motion_notify_event", 
  //  G_CALLBACK (MotionNotifyEvent), (gpointer)psFunctyData);
  g_signal_connect (G_OBJECT (psFunctyData->psDrawingArea), "scroll_event", 
    G_CALLBACK (ScrollEvent), (gpointer)psFunctyData);

  g_signal_connect (G_OBJECT (psFunctyData->psDrawingArea), "map_event", G_CALLBACK (MapEvent), 
    (gpointer)psFunctyData);
  g_signal_connect (G_OBJECT (psFunctyData->psDrawingArea), "unmap_event", G_CALLBACK (UnmapEvent), 
    (gpointer)psFunctyData);
  g_signal_connect (G_OBJECT (psFunctyData->psDrawingArea), "visibility_notify_event", 
    G_CALLBACK (VisibilityNotifyEvent), (gpointer)psFunctyData);

  g_signal_connect (G_OBJECT (psWindow), "key_press_event", G_CALLBACK (KeyPressEvent), 
    (gpointer)psFunctyData);
  g_signal_connect (G_OBJECT (psWindow), "key_release_event", G_CALLBACK (KeyReleaseEvent), 
    (gpointer)psFunctyData);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "PlotApply");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Plot");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (PlotFunctionPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "EditFunction");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (EditFunctionPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "AddCartesian");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (AddCartesianPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "AddSpherical");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (AddSphericalPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "AddCurve");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (AddCurvePress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "DeleteFunction");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (DeleteFunctionPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "OpenView");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (ViewPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "OpenRange");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (RangePress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "RangeApply");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "RangeOkay");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (RangeSetPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Open");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (LoadFilePress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Save");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (SaveFilePress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Export");
  g_signal_connect (psWidget, "clicked", G_CALLBACK (ExportFilePress), (gpointer)psFunctyData);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "FunctionEdit");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "RadiusEdit");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "CurveXEdit");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "CurveYEdit");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "CurveZEdit");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "AccuracyCurve");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "AccuracyRadiusCurve");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Red");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Green");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Blue");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Alpha");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Accuracy");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XCentre");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YCentre");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZCentre");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexFile");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXScale");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYScale");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexXOffset");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "TexYOffset");
  g_signal_connect (psWidget, "activate", G_CALLBACK (PlotFunctionApply), (gpointer)psFunctyData);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Spin");
  g_signal_connect (psWidget, "toggled", G_CALLBACK (ToggleSpinPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Invert");
  g_signal_connect (psWidget, "toggled", G_CALLBACK (ToggleInvertPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Fullscreen");
  g_signal_connect (psWidget, "toggled", G_CALLBACK (ToggleFullscreenPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "DrawAxes");
  g_signal_connect (psWidget, "toggled", G_CALLBACK (ToggleDrawAxesPress), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Wireframe");
  g_signal_connect (psWidget, "toggled", G_CALLBACK (ToggleWireframePress), (gpointer)psFunctyData);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XMin");
  g_signal_connect (psWidget, "key-release-event", G_CALLBACK (RangeMinChange), (gpointer)psFunctyData);
  g_signal_connect (psWidget, "activate", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YMin");
  g_signal_connect (psWidget, "key-release-event", G_CALLBACK (RangeMinChange), (gpointer)psFunctyData);
  g_signal_connect (psWidget, "activate", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZMin");
  g_signal_connect (psWidget, "key-release-event", G_CALLBACK (RangeMinChange), (gpointer)psFunctyData);
  g_signal_connect (psWidget, "activate", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XWidth");
  g_signal_connect (psWidget, "key-release-event", G_CALLBACK (RangeMinChange), (gpointer)psFunctyData);
  g_signal_connect (psWidget, "activate", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YWidth");
  g_signal_connect (psWidget, "key-release-event", G_CALLBACK (RangeMinChange), (gpointer)psFunctyData);
  g_signal_connect (psWidget, "activate", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZWidth");
  g_signal_connect (psWidget, "key-release-event", G_CALLBACK (RangeMinChange), (gpointer)psFunctyData);
  g_signal_connect (psWidget, "activate", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "XMax");
  g_signal_connect (psWidget, "key-release-event", G_CALLBACK (RangeMaxChange), (gpointer)psFunctyData);
  g_signal_connect (psWidget, "activate", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "YMax");
  g_signal_connect (psWidget, "key-release-event", G_CALLBACK (RangeMaxChange), (gpointer)psFunctyData);
  g_signal_connect (psWidget, "activate", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "ZMax");
  g_signal_connect (psWidget, "key-release-event", G_CALLBACK (RangeMaxChange), (gpointer)psFunctyData);
  g_signal_connect (psWidget, "activate", G_CALLBACK (RangeSetApply), (gpointer)psFunctyData);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "Colour");
  g_signal_connect (psWidget, "color-set", G_CALLBACK (SetColourButtonCallback), (gpointer)psFunctyData);

  // Function list
  psFunctionListModel = gtk_list_store_new (FUNCSCOL_NUM, G_TYPE_POINTER, G_TYPE_STRING);
  PopulateFunctionList (psFunctionListModel, psFunctyData);

  psWidget = glade_xml_get_widget (psFunctyData->psXML, "FunctionList");
  gtk_tree_view_set_model (GTK_TREE_VIEW (psWidget), GTK_TREE_MODEL (psFunctionListModel));

  psListCellRenderer = gtk_cell_renderer_text_new ();
  psListColumn = gtk_tree_view_column_new_with_attributes ("Function", 
    psListCellRenderer, "text", FUNCSCOL_FUNCTION, NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (psWidget), psListColumn);

  gtk_tree_view_column_set_cell_data_func (psListColumn, psListCellRenderer, FunctionCellDataFunc, NULL, NULL);

  g_signal_connect (GTK_TREE_VIEW (psWidget), "row-activated", 
    G_CALLBACK (EditFunctionRowDoubleClick), (gpointer)psFunctyData);

  // Display the window
  gtk_widget_show (psWindow);

  // Initialise visualisation
  Init (psVisData);

  // Load the configuration file if there is one
  ClearAllFunctions (psFunctyData);
  LoadFile (DEFAULT_SETTINGS_FILE, TRUE, psFunctyData);
  PopulateFunctionList (psFunctionListModel, psFunctyData);

  // Main loop
  gtk_main ();

  // Save out the configuration file
  SaveFile (DEFAULT_SETTINGS_FILE, TRUE, psFunctyData);

  Deinit (psVisData);

  // Delete persistent structures
  psFunctyData->psVisData = NULL;
  DeleteFunctyPersist (psFunctyData);

  return 0;
}

gchar * GetTitleFilename (char const * szFilename) {
  char * szDisplayName;

  szDisplayName = g_filename_display_basename (szFilename);

  return szDisplayName;
}

void SetMainWindowTitle (char const * szFilename, FunctyPersist * psFunctyData) {
  GtkWidget * psWidget;
  GString * szTitle;
  gchar * szDisplayName;

  // Figure out what to call the window
  szDisplayName = GetTitleFilename (szFilename);
  if (strlen (szDisplayName) > 0) {
    szTitle = g_string_new ("");
    g_string_printf (szTitle, "%s - Functy", szDisplayName);
  }
  else {
    szTitle = g_string_new ("Functy");
  }

  // Set the window title
  psWidget = glade_xml_get_widget (psFunctyData->psXML, "MainWindow");
  gtk_window_set_title (GTK_WINDOW (psWidget), szTitle->str);

  g_free (szDisplayName);
  g_string_free (szTitle, TRUE);
}

void SetExportBinary (bool boBinary, FunctyPersist * psFunctyData) {
	psFunctyData->boBinary = boBinary;
}

void SetExportMultiplier (double fExportMultiplier, FunctyPersist * psFunctyData) {
	psFunctyData->fExportMultiplier = fExportMultiplier;
}

void SetExportScreen (bool boExportScreen, FunctyPersist * psFunctyData) {
	psFunctyData->boExportScreen = boExportScreen;
}

void SetExportAlpha (bool boExportAlpha, FunctyPersist * psFunctyData) {
	psFunctyData->boExportAlpha = boExportAlpha;
}


